import numpy as np

from .utils import gen_rot
from .exceptions import NotOnSurfaceError, InvalidGeometry
from .index import glass_index

from typing import Dict, List, Tuple, Union, Optional

class geometry:
    """Class for the different surfaces in an optical system.

    Attributes
    ----------
    P : list of 3 floats/ints or just float/int
        Position of surface in the lab frame. If only one float or int is given
        then it is considered the position in the Z direction.
    D (optional): list of 3 floats/ints
        Rotation angles alpha, beta, gamma for a surface.
        Defaults to no rotation -> [0., 0., 0.]
    action : str
        Interaction of rays with the specified surface. Either reflection,
        refraction, or stop (do nothing).
    Diam : float/int
        Outer diameter of the specified surface.
    diam : float/int
        Inner diameter of the specified surface.
    N (optional): float/int
        Index of refraction of the material that the rays propagate INTO
        (meaning on the image side). If N is not given then it defaults
        to a value of 1 for air. A constant value can be specified such as 1.72.
    kappa (optional): float/int
        Defines the type of conic that the surface models.
        If kappa is None then the surface is planar.
        kappa < 0 -> hyperboloid
        kappa = 0 -> paraboloid
        0 < kappa < 1 -> helioid of revolution about major axis
        kappa = 1 -> hemisphere
        kappa > 1 -> helioid of revolution about minor axis
    c (optional): float/int
        Vertex curvature of the surface.
        If c is 0 then the surface is planar.
    name (optional): str
        Name of the surface, used for optimization
    R (generated): np.array((3,3))
        Rotation matrix for the surface from rotation angles D.

    """

    def __init__(self, params: Dict):
        P = params.get('P')
        # Allow on axis integer for P.
        if isinstance(P, float) or isinstance(P, int):
            P = np.array([0., 0., P])
        elif isinstance(P, List):
            P = np.array(P)
        elif not isinstance(P, np.ndarray):
            raise InvalidGeometry()
        self.P: np.ndarray = P
        self.D: np.ndarray = np.array(params.get('D', [0., 0., 0.]))
        self.action: str = params['action']
        self.Diam: Union[float, int] = params['Diam']
        self.N: Union[float, int] = params.get('N', 1.)
        self.kappa: Optional[Union[float, int]] = params.get('kappa')
        self.diam: Union[float, int] = params.get('diam', 0.)
        self.c: Union[float, int] = params.get('c', 0.)
        self.name: str = params.get('name', None)
        self.R: np.ndarray = gen_rot(self.D)
        if params.get('glass'):
            self.glass = glass_index(params.get('glass'))
        self.check_params()

    def check_params(self) -> None:
        """Check that required parameters are given and update needed parameters.

        Summary
        -------
            If c != 0 (in the case of a conic) then kappa must be specified,
            and if kappa is greater than 0 then the value of c is redundant
            by boundary conditions of the conic equation. Lastly, if c == 0 in
            the case of a planar surface the None value of kappa needs to be set
            to a dummy value to avoid exceptions in calculating the conic equation.
            Note that this does not affect the calculation since c is 0.

        """
        if self.c != 0:
            if self.kappa is None:
                raise Exception("Specify a kappa for this conic.")
            elif self.kappa > 0:
                print("Warning: Specified c value is not used when kappa>0")
                self.c = np.sqrt(1 / (self.kappa * pow(self.Diam / 2., 2)))
        elif self.c == 0 and self.kappa is None:
            # Used for planes, does not affect calculations.
            self.kappa = 1.

    def get_surface(self, point: np.ndarray) -> Tuple[float, List[float]]:
        """ Returns the function and derivative of a surface for a point. """
        return self.conics(point)

    def get_surface_plot(self, points: np.ndarray) -> np.ndarray:
        """ Returns the function value for an array of points. """
        return self.conics_plot(points)

    def get_surface_vector(self, points: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Returns function values and derivative vectors for an array of points.

        Parameters
        ----------
        points : np.ndarray of shape (N,3)
            Array of points.
        Returns
        -------
        func : np.ndarray of shape (N,)
            Function values.
        deriv : np.ndarray of shape (N,3)
            Derivative vectors.
        """
        return self.conics_vector(points)

    def conics(self, point: np.ndarray) -> Tuple[float, List[float]]:
        """Returns function value and derivative list for conics and sphere surfaces.

        Note
        ----
        This differs from the conics_plot method because we are returning the function
        value from the conic at the point X,Y,Z rather than the Z value for a given X, Y
        pair. A surface is defined by the equation F(X, Y, Z) = 0, so if the ray
        intersects with the surface this function will return a function value of
        0.

        Parameters
        ----------
        point : np.array of length 3
            X, Y, and Z values for the point.

        Returns
        -------
        function : float
            Function value of the surface. A value of 0 corresponds to the ray intersecting
            the surface.
        derivative : list of length 3
            Derivative of the surface at the point given.
        """
        X, Y, Z = point
        rho = np.sqrt(pow(X, 2) + pow(Y, 2))
        if rho > self.Diam / 2. or rho < self.diam / 2.:
            raise NotOnSurfaceError()
        if self.kappa is None:
            raise ValueError("kappa must not be None for conic calculations")
        function = Z - self.c * pow(rho, 2) / (1 + pow((1 - self.kappa * pow(self.c, 2) * pow(rho, 2)), 0.5))
        E = self.c / pow((1 - self.kappa * pow(self.c, 2) * pow(rho, 2)), 0.5)
        derivative = [-X * E, -Y * E, 1.]
        return function, derivative

    def conics_vector(self, points: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Returns function values and derivative vectors for conics and sphere surfaces.
        
        Vectorized version for an array of points.
        
        Parameters
        ----------
        points : np.ndarray of shape (N,3)
            Array of points (X, Y, Z).
        
        Returns
        -------
        func : np.ndarray of shape (N,)
            Function values for each point.
        deriv : np.ndarray of shape (N,3)
            Derivative vectors for each point.
        """
        X = points[:, 0]
        Y = points[:, 1]
        Z = points[:, 2]
        rho = np.sqrt(X**2 + Y**2)
        # Initialize outputs
        func = np.full(points.shape[0], np.nan)
        deriv = np.full((points.shape[0], 3), np.nan)
        # Determine valid indices based on aperture
        valid = (rho <= self.Diam / 2.) & (rho >= self.diam / 2.)
        if self.kappa is None:
            raise ValueError("kappa must not be None for conic calculations")
        # Calculate for valid points
        rho_valid = rho[valid]
        sqrt_term = np.sqrt(1 - self.kappa * self.c**2 * rho_valid**2)
        denom = 1 + sqrt_term
        func[valid] = Z[valid] - self.c * rho_valid**2 / denom
        E = self.c / sqrt_term
        deriv[valid, 0] = -X[valid] * E
        deriv[valid, 1] = -Y[valid] * E
        deriv[valid, 2] = 1.
        return func, deriv

    def conics_plot(self, point: np.ndarray) -> np.ndarray:
        """Returns Z values for an array of points for plotting conics.

        Parameters
        ----------
        point : 2d np.array with shape (n, 3)
            X, Y, and Z values for each point (for each row).

        Returns
        -------
        function : 1d np.array
            np.array of Z values for each X,Y pair.
        """
        X, Y = point[:, 0], point[:, 1]
        rho = np.sqrt(pow(X, 2) + pow(Y, 2))
        function = np.zeros(len(point))
        nan_idx = (rho > self.Diam / 2.) + (rho < self.diam / 2.)
        rho = np.sqrt(pow(X[~nan_idx], 2) + pow(Y[~nan_idx], 2))
        function[nan_idx] = np.nan
        if self.kappa is None:
            raise ValueError("kappa must not be None for conic plot calculations")
        function[~nan_idx] = self.c * pow(rho, 2) / (1 + pow((1 - self.kappa * pow(self.c, 2) * pow(rho, 2)), 0.5))
        return function


"""
Ray tracing and optical design module for Python
================================================

TracePy is a sequential ray tracing package written
in Python for designing optical systems in the geometric
optics regime. It features lens optimization from Scipy.
TracePy is currently in active development and any collaborators
would be welcome.
"""

from .ray import RayGroup
from .geometry import geometry
from .optimize import optimize
from .geoplot import plotxz, plotyz, plot2d
from .optplot import spotdiagram, plotobject, rayaberration, spot_rms
from .iotables import save_optics
from .raygroup import ray_plane
from .index import cauchy_two_term, glass_index
from .utils import gen_rot


class NormalizationError(Exception):
    """ Custom exception for unnormalized input. """

class NotOnSurfaceError(Exception):
    """ Error for rays that do not intersect with a surface. """

class TraceError(Exception):
    """ Custom error for lens systems where no rays survive being traced. """

class InvalidGeometry(Exception):
    """ Invalid parameters were given to define a geometry object. """



import numpy as np
from pytest import approx

from tracepy import geometry

surface1 = {
    'P': 2.,
    'D': [0., 0., 2.],
    'action': 'reflection',
    'Diam': 2.,
    'diam': 1.,
    'N': 2.,
    'kappa': -2.,
    'c': 3.,
    'name': 'hello',
}

surface2 = {
    'P': [2., 3., 4.],
    'action': 'stop',
    'Diam': 2.,
    'glass': 'F2 schott',
    'c': -3,
    'kappa': 0.,
}

geo = [surface1, surface2]
geo_obj = [geometry(surf) for surf in geo]

def test_params():
    """ Read parameters to make sure nothing was changed. """
    print(geo_obj[0].P)
    assert np.all(geo_obj[0].P == [0., 0., 2.])
    assert np.all(geo_obj[0].D == [0., 0., 2.])
    assert geo_obj[0].action == 'reflection'
    assert geo_obj[0].Diam == 2.
    assert geo_obj[0].diam == 1.
    assert geo_obj[0].N == 2.
    assert geo_obj[0].kappa == -2.
    assert geo_obj[0].c == 3.
    assert geo_obj[0].name == 'hello'

    assert np.all(geo_obj[1].P == [2., 3., 4.])
    assert np.all(geo_obj[1].D == [0., 0., 0.])
    assert geo_obj[1].action == 'stop'
    assert geo_obj[1].Diam == 2.
    assert geo_obj[1].kappa == 0.
    assert geo_obj[1].c == -3.

def test_conics():
    """ Check simple function cases. """
    assert geo_obj[1].conics(np.array([0., 0., 0.]))[0] == 0.0
    assert np.all(geo_obj[1].conics(np.array([0., 0., 0.]))[1] == [0., 0., 1.])

    assert approx(geo_obj[0].conics(np.array([1., 0., 0.]))[0], 0.00002) == -0.55982
    assert approx(geo_obj[0].conics(np.array([1., 0., 0.]))[1][0], 0.00002) == -0.68825
    assert geo_obj[0].conics(np.array([1., 0., 0.]))[1][1] == -0.0
    assert geo_obj[0].conics(np.array([1., 0., 0.]))[1][2] == 1.0

def test_conics_plot():
    """ Check simple Z value returns. """
    assert geo_obj[1].conics_plot(np.array([[0., 0.]]))[0] == 0.



import numpy as np
from scipy.optimize import least_squares

from .optplot import spot_rms
from .raygroup import ray_plane
from .constants import SURV_CONST, MAX_RMS
from .exceptions import TraceError

from typing import List, Union, Dict, Optional

def update_geometry(inputs: List[Union[float, int]],
                    geoparams: List[Dict],
                    vary_dicts: List[Dict]) -> List[Dict]:
    """Return the geometry requested by the optimization algorithm.

    Parameters
    ----------
    inputs : list of floats/ints
        Values for varying parameters that the optimizer selected.
    geoparams : list of dictionaries
        Surface parameters in dictionary form.
    vary_dicts : list of dictionaries
        Lists the parameters that the user wants to optimize.

    Returns
    -------
    geoparams : list of dictionaries
        Returns the list of paramaters after updating it with inputs.

    """

    vary_idxs = 0
    for dict_ in vary_dicts:
        name = dict_["name"]
        vary_list = dict_["vary"]
        for surface in geoparams:
            if "name" in surface and surface["name"] == name:
                for item in vary_list:
                    surface[item] = inputs[vary_idxs]
                    vary_idxs += 1
    return geoparams

def get_rms(inputs: List[Union[float, int]],
            geoparams: List[Dict],
            vary_dicts: List[Dict]) -> float:
    """Return the RMS of an updated geometry.

    Note
    ----
    If no rays survive then a large RMS value is chosen so that the optimizer is discouraged from looking
    there.

    Parameters
    ----------
    inputs : list of floats/ints
        Values for varying parameters that the optimizer selected.
    geoparams : list of dictionaries
        Surface parameters in dictionary form.
    vary_dicts : list of dictionaries
        Lists the parameters that the user wants to optimize.

    Returns
    -------
    rms : float
        RMS of the spotdiagram.
    """
    params_iter = update_geometry(inputs, geoparams, vary_dicts)
    raygroup_iter = ray_plane(params_iter, [0., 0., 0.], 1.1, d=[0., 0., 1.], nrays=50)
    rays_list = raygroup_iter.to_ray_list()
    ratio_surv = np.sum([1 for ray in rays_list if ray.active]) / len(rays_list)
    try:
        rms = spot_rms(params_iter, rays_list)
    except TraceError:
        rms = MAX_RMS
    return rms + (1 - ratio_surv) * SURV_CONST

def optimize(geoparams:List[Dict],
             vary_dicts: List[Dict],
             typeof: str = 'least_squares',
             max_iter: Optional[int] = None) -> List[Dict]:
    """Optimize a given geometry for a given varylist and return the new geometry.

    Parameters
    ----------
    geoparams : list of dictionaries
        Surface parameters in dictionary form.
    vary_dicts : list of dictionaries
        Lists the parameters that the user wants to optimize.
    typeof : str
        Type of optimization to use.
    max_iter : int
        Max number of function calls until the optimizer stops.

    Returns
    -------
    new_params : list of dictionaries
        List of otimized dictionaries that describe surfaces.

    """

    initial_guess = []
    param_lb = []
    param_ub = []
    for dict_ in vary_dicts:
        name = dict_["name"]
        vary_list = dict_["vary"]
        for surface in geoparams:
            if "name" in surface and surface["name"] == name:
                for item in vary_list:
                    attr = surface[item]
                    if item == 'P' and (isinstance(item, float) or isinstance(item, int)):
                         attr = attr[2]
                    if item == 'c' or item == 'kappa':
                        lower_bound = 0 if attr > 0 else -np.inf
                        upper_bound = np.inf if attr >= 0 else 0.
                        param_lb.append(lower_bound)
                        param_ub.append(upper_bound)
                    else:
                        param_lb.append(-np.inf)
                        param_ub.append(np.inf)
                    initial_guess.append(attr)
    param_bounds = [tuple(param_lb), tuple(param_ub)]
    if typeof == 'least_squares':
        res = least_squares(get_rms, initial_guess, args=(geoparams, vary_dicts), max_nfev=max_iter,\
                            bounds = param_bounds)
    new_params = update_geometry(np.around(res.x, 3), geoparams, vary_dicts)
    return new_params



import numpy as np

from .geometry import geometry

from typing import Optional, Tuple, Union

def transform_points(R: np.ndarray,
                     surface: geometry,
                     points: np.ndarray) -> np.ndarray:
    """Transforms points into the reference frame of surface.

    Note
    ----
    Arrays are flattened before return if they only have one row.

    Parameters
    ----------
    R : np.array((3,3))
        Rotation matrix for surface.
    surface : geometry object
        Surface whose reference frame to transform into.
    points : 2d np.array
        Points for each row that will be transformed.

    Returns
    -------
    tran_points : 2d np.array
        Points in the transformed reference frame.
    """
    points = np.atleast_2d(points)
    tran_points = np.dot(R, (points - surface.P).T).T
    if tran_points.shape[0] == 1:
        tran_points = tran_points.flatten()
    return tran_points

def transform_dir(R: np.ndarray,
                  surface: geometry,
                  D: np.ndarray) -> np.ndarray:
    """Transforms directions into the reference frame of surface.

    Note
    ----
    Arrays are flattened before return if they only have one row.

    Parameters
    ----------
    R : np.array((3,3))
        Rotation matrix for surface.
    surface : geometry object
        Surface whose reference frame to transform into.
    D : 2d np.array
        Directions for each row that will be transformed.

    Returns
    -------
    tran_D : 2d np.array
        Directions in the transformed reference frame.
    """
    D = np.atleast_2d(D)
    tran_D = np.dot(R, D.T).T
    if tran_D.shape[0] == 1:
        tran_D = tran_D.flatten()
    return tran_D

def lab_frame_points(R: np.ndarray,
                     surface: geometry,
                     points: np.ndarray) -> np.ndarray:
    """Transforms points into the lab frame.

    Note
    ----
    Arrays are flattened before return if they only have one row.

    Parameters
    ----------
    R : np.array((3,3))
        Rotation matrix for surface.
    surface : geometry object
        Surface whose reference frame to transform from.
    points : 2d np.array
        Points for each row that will be transformed.

    Returns
    -------
    lab_points : 2d np.array
        Points in the lab reference frame.
    """
    points = np.atleast_2d(points)
    lab_points = np.dot(R.T, points.T).T + surface.P
    if lab_points.shape[0] == 1:
        lab_points = lab_points.flatten()
    return lab_points

def lab_frame_dir(R: np.ndarray,
                  surface: geometry,
                  D: np.ndarray) -> np.ndarray:
    """Transforms directions into the lab frame.

    Note
    ----
    Arrays are flattened before return if they only have one row.

    Parameters
    ----------
    R : np.array((3,3))
        Rotation matrix for surface.
    surface : geometry object
        Surface whose reference frame to transform from.
    D : 2d np.array
        Directions for each row that will be transformed.

    Returns
    -------
    lab_D : 2d np.array
        Directions in the lab reference frame.
    """
    D = np.atleast_2d(D)
    lab_D = np.dot(R.T, D.T).T
    if lab_D.shape[0] == 1:
        lab_D = lab_D.flatten()
    return lab_D


import numpy as np
import tracepy as tp

mirror = {
    'action': 'reflection',
    'D': np.array([0., np.pi, 0.]),
    'P': np.array([0., 0., 10.]),
    'kappa': 0.,
    'c': 0.05,
    'Diam': 3.6,
}

small_mirror = {
    'action': 'reflection',
    'D': np.array([0., -np.pi/4., 0.]),
    'P': np.array([0., 0., 2.]),
    'Diam': 1.4,
}

selector = {
    'action': 'stop',
    'P': np.array([0., 0., 0.1]),
    'Diam': 3.6,
    'diam': 1.4
}

stop = {
    'action': 'stop',
    'P': np.array([0., -4.6, 2.]),
    'D': np.array([0., np.pi/2., 0.]),
    'Diam': 0.8
}

lensy = -4.
thickness = 0.1

lens1 = {
    'action': 'refraction',
    'P': np.array([0., lensy, 2.]),
    'D': np.array([0., 3*np.pi/2., 0.]),
    'Diam': 0.8,
    'kappa': 0.,
    'glass': 'ZF4 cdgm',
    'c': 0.3
}

lens2 = {
    'action': 'refraction',
    'P': np.array([0., lensy - thickness, 2.]),
    'D': np.array([0., 3*np.pi/2., 0.]),
    'Diam': 0.8,
    'kappa': 0.,
    'c': -0.3
}

def test_plotting_functions():
    geo = [selector, mirror, small_mirror, lens1, lens2, stop]
    ray_group = tp.ray_plane(geo, [0., 0., 0.], 1.8, d=[0., 0., 1.], nrays=50, wvl=0.55)

    # Test plotyz
    try:
        tp.plotyz(geo, ray_group)
    except Exception as e:
        assert False, f"plotyz raised an exception: {e}"

    # Test plotxz
    try:
        tp.plotxz(geo, ray_group)
    except Exception as e:
        assert False, f"plotxz raised an exception: {e}"

    # Test spotdiagram
    try:
        tp.spotdiagram(geo, ray_group)
    except Exception as e:
        assert False, f"spotdiagram raised an exception: {e}"

    # Test plotobject
    try:
        tp.plotobject(geo, ray_group)
    except Exception as e:
        assert False, f"plotobject raised an exception: {e}"


import numpy as np
import matplotlib.pyplot as plt

from .geometry import geometry
from .transforms import transform_points
from .constants import PLOT_ROUNDING_ACC
from .exceptions import TraceError

from typing import List, Dict, Tuple, Any

def _gen_object_points(surface: geometry,
                       surface_idx: int,
                       rays: List[Any]) -> np.ndarray:
    """Transform intersection points into a surface's reference frame.

    Parameters
    ----------
    surface : geometry object
        Surface whose reference frame the points will be transformed into.
    surface_idx : int
        Integer corresponding to where the surface is in the propagation
        order. For example, 0 means the surface is the first surface rays
        are propagated through.
    rays : list of ray objects
        Rays that were propagated through the geometry list.

    Returns
    -------
    points_obj: 2d np.array
        X, Y pair points in 2d array for easy RMS calculation.
    """
    # Gather the history points for all active rays.
    points = np.array([rayiter.P_hist[surface_idx] for rayiter in rays if rayiter.active])
    if points.size == 0:
        # Instead of raising an error, return a single point at (0,0).
        return np.array([[0.0, 0.0]])
    # Transform to the object (surface) frame.
    points_obj = transform_points(surface.R, surface, points)
    points_obj = np.around(points_obj, PLOT_ROUNDING_ACC)
    # Filter out any rows that contain NaN.
    valid = ~np.isnan(points_obj).any(axis=1)
    if np.sum(valid) == 0:
        return np.array([[0.0, 0.0]])
    return points_obj[valid]

def calculate_rms(points: np.ndarray) -> float:
    """Calculates the RMS of the given points.

    Parameters
    ----------
    points : np.ndarray
        Array of points to calculate RMS for.

    Returns
    -------
    float
        The calculated RMS value.
    """
    # If only one valid point exists, the RMS is zero.
    if points.shape[0] < 2:
        return 0.0
    return np.std(points[:, [0, 1]] - points[:, [0, 1]].mean(axis=0))

def spot_rms(geo_params: List[Dict], rays: Any) -> float:
    """Calculates the RMS of the spot diagram points.

    Parameters
    ----------
    geo_params : list of dictionaries
        Surfaces in order of propagation.
    rays : list of ray objects or RayGroup object
        Rays that were propagated through the geometry list. If a RayGroup
        object is provided, it will be converted automatically.

    Returns
    -------
    float
        The calculated RMS value.
    """
    if hasattr(rays, "to_ray_list"):
        rays = rays.to_ray_list()
    stop = geometry(geo_params[-1])
    points_obj = _gen_object_points(stop, -1, rays)
    return calculate_rms(points_obj)

def spotdiagram(geo_params: List[Dict],
                rays: Any,
                pltparams: Dict = {'color': 'red'}) -> None:
    """Plots the transformed intersection points of rays on the stop surface.

    Parameters
    ----------
    geo_params : list of dictionaries
        Surfaces in order of propagation.
    rays : list of ray objects or RayGroup object
        Rays that were propagated through the geometry list. If a RayGroup
        object is provided, it will be converted automatically.
    pltparams : dictionary
        Plotting attributes of the spot diagram.
    """
    if hasattr(rays, "to_ray_list"):
        rays = rays.to_ray_list()
    stop = geometry(geo_params[-1])
    points_obj = _gen_object_points(stop, -1, rays)
    X, Y = points_obj[:, 0], points_obj[:, 1]
    rms = calculate_rms(points_obj)

    plt.subplot(1, 1, 1, aspect='equal')
    plt.locator_params(axis='x', nbins=8)
    plt.ticklabel_format(style='sci', axis='both', scilimits=(0, 0))
    plt.plot(X, Y, '+', **pltparams)
    plt.text(0.65, 1.015, 'RMS=%.2E' % rms, transform=plt.gca().transAxes)

def plotobject(geo_params: List[Dict],
               rays: Any,
               pltparams: Dict = {'color': 'blue'}) -> None:
    """Plots the initial ray locations in the initial object's frame.

    Note
    ----
    Directly plots the object spot diagram rather than returning something.

    Parameters
    ----------
    geo_params : list of dictionaries
        Surfaces in order of propagation.
    rays : list of ray objects or RayGroup object
        Rays that were propagated through the geometry list. If a RayGroup
        object is provided, it will be converted automatically.
    pltparams : dictionary
        Plotting attributes of the object spot diagram.
    """
    if hasattr(rays, "to_ray_list"):
        rays = rays.to_ray_list()
    start = geometry(geo_params[0])
    points_obj = _gen_object_points(start, 0, rays)
    X, Y = points_obj[:, 0], points_obj[:, 1]
    rms = np.std(points_obj[:, [0, 1]] - points_obj[:, [0, 1]].mean(axis=0))
    plt.subplot(1, 1, 1, aspect='equal')
    plt.ticklabel_format(style='sci', axis='both', scilimits=(0, 0))
    plt.text(0.65, 1.015, 'RMS=%.2E' % rms, transform=plt.gca().transAxes)
    plt.plot(X, Y, '+', **pltparams)

def rayaberration(geo_params: List[Dict],
                  rays: Any,
                  pltparams: Dict = {'color': 'red', 'linewidth': 2}) -> None:
    """Plots the ray aberration diagrams for an optical system.

    Note
    ----
    Directly plots the ray aberration curve rather than returning something.

    Parameters
    ----------
    geo_params : list of dictionaries
        Surfaces in order of propagation.
    rays : list of ray objects or RayGroup object
        Rays that were propagated through the geometry list. If a RayGroup
        object is provided, it will be converted automatically.
    pltparams : dictionary
        Plotting attributes of the curve.
    """
    if hasattr(rays, "to_ray_list"):
        rays = rays.to_ray_list()
    start = geometry(geo_params[0])
    start_points = _gen_object_points(start, 0, rays)
    X_start, Y_start = start_points[:, 0], start_points[:, 1]
    stop = geometry(geo_params[-1])
    stop_points = _gen_object_points(stop, -1, rays)
    X_stop, Y_stop = stop_points[:, 0], stop_points[:, 1]
    # Chief ray information
    distances = np.sqrt(np.sum((start_points[:, [0, 1]] - start_points[:, [0, 1]].mean(axis=0))**2, axis=1))
    chief_ray_idx = np.argmin(distances)
    chief_ray = start_points[:, [0, 1]][chief_ray_idx]
    ray_subset_idxs = np.where(X_start == chief_ray[0])
    rel_stops = Y_stop[ray_subset_idxs] - Y_stop[chief_ray_idx]
    # Normalized pupil coordinates
    norm_coords = Y_start[ray_subset_idxs] - chief_ray[1]
    norm_coords[np.where(norm_coords < 0)] /= abs(np.min(norm_coords))
    norm_coords[np.where(norm_coords > 0)] /= abs(np.max(norm_coords))
    # Used to show axes when there is no aberration
    if np.all(rel_stops == 0.):
        plt.plot([0., 0.], [-0.04, 0.04], 'k')
    else:
        plt.plot([0] * len(rel_stops), rel_stops, 'k')
    plt.plot(norm_coords, [0] * len(norm_coords), 'k')
    plt.plot(norm_coords, rel_stops, **pltparams)


# Constants used by optimizer
SURV_CONST = 100 # Weight of failed propagation.
MAX_RMS = 999 # Maximum RMS penalty for trace error.

# Constants used for ray objects
MAX_INTERSECTION_ITERATIONS = 1e4 # Max iter before failed intersection search.
MAX_REFRACTION_ITERATIONS = 1e5 # Max iter before failed refraction.
INTERSECTION_CONVERGENCE_TOLERANCE = 1e-6 # Tolerance for intersection search.
REFRACTION_CONVERGENCE_TOLERANCE = 1e-15 # Tolerance for refraction.

# Constants used for plotting
PLOT_ROUNDING_ACC = 14 # Rounding accuracy for spot diagrams and plots.


import tracepy as tp

mirror = {
    'action': 'reflection',
    'P': 1.5,
    'kappa': 0.,
    'c': -0.5,
    'Diam': 2.2,
}

stop = {
    'action': 'stop',
    'P': 0.5,
    'Diam': 0.2
}

def test_rms_parabolic():
    geo = [mirror, stop]
    ray_group = tp.ray_plane(geo, [0., 0., -1.5], 1.1, d=[0., 0., 1.], nrays=100)
    rms = tp.spot_rms(geo, ray_group)
    assert rms == 0.


import os
import json
import yaml

import numpy as np
from scipy.optimize import curve_fit

def cauchy_two_term(x: float, B: float, C: float) -> float:
    """
    This is a simple two-term Cauchy equation for the index of refraction as a function of wavelength.
    https://en.wikipedia.org/wiki/Cauchy%27s_equation
    It is used to create a function for refractive index database entries when only tabulated data is available.
    
    Parameters
    ----------
    x : float
        Wavelength (in micrometers).
    B : float
        First term of the Cauchy equation.
    C : float
        Second term of the Cauchy equation.
        
    Returns
    -------
    float
        Refractive index at the given wavelength.
    """
    return B + (C / (x ** 2))

# TODO: This function, written by @MikeMork, needs to be broken up into several smaller functions with explicit typing.
def glass_index(glass):
    '''
    Given a glass name this function will return a function that takes wavelength in microns and returns index of refraction.
    - All values were taken from refractiveindexinfo's database of different optical glasses.

    https://refractiveindex.info/

    A constant value can also be input to this function to allow directly setting index

    Some glass names like F2 are shared by multiple manufacturers.
    Example input specifying the manufacturer: "F2 Schott"
    Example input without manufacturer: "F2"

    If manufacturer is not specified the following priority are used to choose the glass:
    1)schott
    2)ohara
    3)sumita
    4)cdgm
    5)hoya
    6)hikari
    7)vitron
    8)ami
    9)barberini
    10)corning
    11)lightpath
    12)lzos
    13)misc
    14)nsg

    This list isn't really a comprehensive list of manufacturers but rather the directory structure of the refractiveindexinfo yaml database.

    The functions that this function returns take a wavelength in microns and return the index of refractive index.
    '''
    # Load glass dictionary
    with open(os.path.join(os.path.dirname(__file__), "glass", "glass_dict.json"), "r") as f:
        glass_dict = json.load(f)
    glass_catalog = ["schott", "ohara", "sumita", "cdgm", "hoya", "hikari",
                     "vitron", "ami", "barberini", "lightpath", "lzos", "misc", "nsg"]

    # Process input and return glass file:
    if isinstance(glass, (float, int)):
        constant_index = float(glass)
        return lambda x=0.55: constant_index  # scalar; broadcasting works fine
    elif isinstance(glass, str):
        input_list = glass.split()
        glass_name = input_list[0]
        glass_manufacturer = input_list[1].lower() if len(input_list) > 1 else None

        if glass_name in glass_dict:
            if glass_manufacturer:
                if glass_manufacturer in glass_dict[glass_name]:
                    glass_file = glass_dict[glass_name][glass_manufacturer]
                else:
                    raise Exception(f"{glass_name} is not in the glass catalog for {glass_manufacturer}.")
            else:
                available = list(glass_dict[glass_name].keys())
                # choose based on the defined priority order
                priority = sorted(available, key=lambda x: glass_catalog.index(x) if x in glass_catalog else np.inf)
                glass_file = glass_dict[glass_name][priority[0]]
        else:
            raise Exception(f"{glass_name} is not in the glass catalog.")

        # Process glass file and return the correct function (wavelength in microns)
        split_path = glass_file.split("\\")
        if len(split_path) == 4:
            glass_file_path = os.path.join(split_path[1], split_path[2], split_path[3])
        else:
            glass_file_path = os.path.join(split_path[1], split_path[2], split_path[3], split_path[4])
        f = yaml.load(open(os.path.join(os.path.dirname(__file__), glass_file_path)), Loader=yaml.BaseLoader)
        type_index_function = f["DATA"][0]["type"]
        
        # Use np.asarray(x) so that the operations are vectorized
        if type_index_function == 'formula 1':
            coeffs = [float(x) for x in f["DATA"][0]['coefficients'].split()]
            return lambda x=0.55: np.sqrt(
                1 + coeffs[0] +
                coeffs[1] / (1 - (coeffs[2] / np.asarray(x))**2) +
                coeffs[3] / (1 - (coeffs[4] / np.asarray(x))**2)
            )
        elif type_index_function == 'formula 2':
            coeffs = [float(x) for x in f["DATA"][0]['coefficients'].split()]
            if len(coeffs) >= 7:
                return lambda x=0.55: np.sqrt(
                    1 +
                    coeffs[1] / (1 - coeffs[2] / np.asarray(x)**2) +
                    coeffs[3] / (1 - coeffs[4] / np.asarray(x)**2) +
                    coeffs[5] / (1 - coeffs[6] / np.asarray(x)**2)
                )
            else:
                return lambda x=0.55: np.sqrt(
                    1 +
                    coeffs[1] / (1 - coeffs[2] / np.asarray(x)**2) +
                    coeffs[3] / (1 - coeffs[4] / np.asarray(x)**2)
                )
        elif type_index_function == 'formula 3':
            coeffs = [float(x) for x in f["DATA"][0]['coefficients'].split()]
            return lambda x=0.55: np.sqrt(
                coeffs[0] -
                coeffs[1] * np.asarray(x)**coeffs[2] +
                coeffs[3] * np.asarray(x)**coeffs[4] +
                coeffs[5] * np.asarray(x)**coeffs[6] +
                coeffs[7] * np.asarray(x)**coeffs[8] +
                coeffs[9] * np.asarray(x)**coeffs[10]
            )
        elif type_index_function == 'formula 5':
            coeffs = [float(x) for x in f["DATA"][0]['coefficients'].split()]
            return lambda x=0.55: (coeffs[0] -
                                     coeffs[1] * np.asarray(x)**coeffs[2] +
                                     coeffs[3] * np.asarray(x)**(-coeffs[4]))
        elif type_index_function in ['tabulated n', 'tabulated nk']:
            raw_data = f["DATA"][0]['data'].strip().split('\n')
            wavelength = np.array([float(line.split()[0]) for line in raw_data])
            index = np.array([float(line.split()[1]) for line in raw_data])
            popt, _ = curve_fit(cauchy_two_term, wavelength, index)
            return lambda x=0.55: popt[0] + popt[1] / (np.asarray(x)**2)
    else:
        raise TypeError("Glass must be either a float/int or a string.")


import numpy as np
import matplotlib.pyplot as plt

from .ray import RayGroup
from .geometry import geometry
from .transforms import lab_frame_points

from typing import List, Dict, Optional, Any

def _gen_points(surfaces: List[geometry]) -> np.ndarray:
    """Generates the mesh points for each surface in the obj frame.

    Parameters
    ----------
    surfaces : list of geometry objects
        Surfaces in propagation order to plot.

    Returns
    -------
    surfpoints : 2d np.array
        Points (X, Y, Z) in the surface's reference frame.
    """
    surfpoints = []
    for surface in surfaces:
        bound = surface.Diam / 2.
        # General mesh points.
        linspace = np.linspace(-bound, bound, 200)
        x_mesh, y_mesh = np.meshgrid(linspace, -linspace)
        # Used for cross sections.
        x_points = np.append(x_mesh, [linspace, np.zeros(200)])
        y_points = np.append(y_mesh, [np.zeros(200), linspace])
        meshpoints_2d = np.vstack((x_points.ravel(), y_points.ravel())).T
        # Function values in obj. frame.
        z_points = surface.get_surface_plot(meshpoints_2d)
        meshpoints = np.vstack((x_points.ravel(), y_points.ravel(), z_points)).T
        # Append to surfpoints which holds all surfaces' points.
        surfpoints.append(meshpoints)
    return np.array(surfpoints)

def _plot_rays(rays: np.ndarray,
               axes: List[int],
               pltparams: Dict) -> None:
    """Plots 2d ray history points. Takes list axes to specify axes (0, 1, 2) to plot.

    Parameters
    ----------
    rays : np.array
        Array of dummy ray objects to be plotted.
    axes : list of length 2 with integers from range [0,2]
        Axes (X, Y, Z) to plot from ray points.
    pltparams : dictionary
        Plot characteristics of rays such as colors and alpha.
    """
    for ray in rays:
        if len(ray.P_hist) < 2:
            continue
        for idx in range(len(ray.P_hist) - 1):
            F, G = ray.P_hist[idx][axes]
            F_p, G_p = ray.P_hist[idx + 1][axes]
            #Alpha, beta, or gamma rotations.
            H_p, I_p = ray.D_hist[idx + 1][axes]
            plt.plot([G, G_p], [F, F_p], **pltparams)
        # Only plot final direction if the loop ran at least once.
        if len(ray.P_hist) >= 2:
            plt.plot([G_p, G_p + I_p], [F_p, F_p + H_p], **pltparams)

def _clip_lens(surfaces: List[geometry], surfpoints: np.ndarray, idx: int) -> np.ndarray:
    """Clips points outside of a lens intersection point.

    Parameters
    ----------
    surfaces : list of geometry objects
        Surfaces whose reference frames to transform from.
    surfpoints : 2d np.array
        Points from surface for each row that will be clipped.
    idx : int
        Index of surface in propagation order.

    Returns
    -------
    surfpoints : 2d np.array
        Points after clipping.
    """
    surf1, surf2 = surfaces[idx], surfaces[idx + 1]
    d = np.sqrt(np.sum(np.square(surf1.P - surf2.P)))
    points1, points2 = np.nan_to_num(surfpoints[idx]), np.nan_to_num(surfpoints[idx + 1])
    points2[:, 2] += d
    clipped_idx = (points2[:, 2] - points1[:, 2]) <= 0.
    surfpoints[idx][:, 2][clipped_idx] = np.nan
    surfpoints[idx + 1][:, 2][clipped_idx] = np.nan
    return surfpoints

def _plot_surfaces(geo_params: List[Dict], axes: List[int]) -> None:
    """Plots 2d surface cross sections. Takes list axes to specify axes (0, 1, 2) to plot surfaces in.

    Note
    ----
    Directly plots the surfaces rather than returning something.

    Parameters
    ----------
    geo_params : list of dictionaries
        Surfaces in propagation order to plot.
    axes : list of length 2 with integers from range [0,2]
        Axes (X, Y, Z) to plot surfaces in.
    """
    surfaces = [geometry(surf) for surf in geo_params]
    surfpoints = _gen_points(surfaces)
    lens_check = 0
    start = None
    for idx, surf in enumerate(surfaces):
        lens_condition = (idx + 1 < len(surfaces) and
                          surfaces[idx].action == surfaces[idx + 1].action == 'refraction')
        if lens_condition:
            surfpoints = _clip_lens(surfaces, surfpoints, idx)
        with np.errstate(invalid='ignore'):
            if np.any(np.mod(surf.D / np.pi, 1) != 0) and surf.c == 0 and surf.diam == 0:
                # Find cross section points.
                cross_idx = abs(surfpoints[idx][:, axes[1]]) == 0
            else:
                # Find cross section points.
                cross_idx = abs(surfpoints[idx][:, 1 - axes[0]]) == 0
        cross_points = surfpoints[idx][cross_idx]
        # Transform to lab frame.
        points = lab_frame_points(surf.R, surf, cross_points)
        F, G = points[:, axes[0]], points[:, axes[1]]
        # Connect the surfaces in a lens
        if surfaces[idx].action == surfaces[idx - 1].action == 'refraction' and start is not None:
            lens_check = 1 - lens_check
            if lens_check == 1:
                start_2 = np.array([F[0], G[0]])
                end_2 = np.array([F[-1], G[-1]])
                dis1 = np.sqrt(np.sum(np.square(start - start_2)))
                dis2 = np.sqrt(np.sum(np.square(start - end_2)))
                if dis1 <= dis2:
                    idxs = [0, -1]
                else:
                    idxs = [-1, 0]
                F = np.insert(F, idxs, [start[0], end[0]])
                G = np.insert(G, idxs, [start[1], end[1]])
        if lens_condition:
            start = np.array([F[0], G[0]])
            end = np.array([F[-1], G[-1]])
        plt.plot(G, F, 'k')

def plotxz(geo_params: List[Dict],
           ray_list: Any,
           pltparams: Dict = {'c': 'red', 'alpha': 0.3},
           both: Optional[bool] = None) -> None:
    """Plots the xz coordinates of all rays and surface cross sections.

    Parameters
    ----------
    geo_params : list of dictionaries
        Surfaces in propagation order to plot.
    ray_list : list or RayGroup object
        Dummy ray objects or RayGroup that are going to be plotted.
    pltparams : dictionary
        Plot characteristics of rays such as colors and alpha.
    both (optional) : bool
        Flag for overriding a 111 subplot for self.plot2d.
    """
    if hasattr(ray_list, "to_ray_list"):
        ray_list = ray_list.to_ray_list()
    rays = np.array([rayiter for rayiter in ray_list if rayiter.active])
    if both is None:
        plt.subplot(1, 1, 1, aspect='equal')
    _plot_rays(rays, [0, 2], pltparams)
    _plot_surfaces(geo_params, axes=[0, 2])
    plt.xlabel("Z")
    plt.ylabel("X")

def plotyz(geo_params: List[Dict],
           ray_list: Any,
           pltparams: Dict = {'c': 'red', 'alpha': 0.3},
           both: Optional[bool] = None) -> None:
    """Plots the yz coordinates of all rays and surface cross sections.

    Parameters
    ----------
    geo_params : list of dictionaries
        Surfaces in propagation order to plot.
    ray_list : list or RayGroup object
        Dummy ray objects or RayGroup that are going to be plotted.
    pltparams : dictionary
        Plot characteristics of rays such as colors and alpha.
    both (optional) : bool
        Flag for overriding a 111 subplot for self.plot2d.
    """
    if hasattr(ray_list, "to_ray_list"):
        ray_list = ray_list.to_ray_list()
    rays = np.array([rayiter for rayiter in ray_list if rayiter.active])
    if both is None:
        plt.subplot(1, 1, 1, aspect='equal')
    _plot_rays(rays, [1, 2], pltparams)
    _plot_surfaces(geo_params, axes=[1, 2])
    plt.xlabel("Z")
    plt.ylabel("Y")

def plot2d(geo_params: List[Dict],
           rays: Any,
           pltparams: Dict = {'c': 'red', 'alpha': 0.3}) -> None:
    """Plots both xz and yz side-by-side.

    Parameters
    ----------
    geo_params : list of dictionaries
        Surfaces in propagation order to plot.
    rays : list or RayGroup object
        Dummy ray objects or RayGroup that are going to be plotted.
    pltparams : dictionary
        Plot characteristics of rays such as colors and alpha.
    """
    if hasattr(rays, "to_ray_list"):
        rays = rays.to_ray_list()
    plt.subplot(2, 1, 1, aspect='equal')
    plotxz(geo_params, rays, pltparams, both=True)
    plt.subplot(2, 1, 2, aspect='equal')
    plotyz(geo_params, rays, pltparams, both=True)
    plt.tight_layout()


import numpy as np

from .transforms import transform_points, transform_dir, lab_frame_points, lab_frame_dir
from .constants import SURV_CONST, MAX_INTERSECTION_ITERATIONS, INTERSECTION_CONVERGENCE_TOLERANCE, MAX_REFRACTION_ITERATIONS, REFRACTION_CONVERGENCE_TOLERANCE
from .exceptions import NormalizationError, NotOnSurfaceError
from .geometry import geometry

from typing import Union, List

class RayGroup:
    """Class for a group of rays and their vectorized propagation through surfaces.

    Attributes
    ----------
    P : np.ndarray of shape (N,3)
        Positions of rays in the lab frame.
    D : np.ndarray of shape (N,3)
        Direction cosines for rays in the lab frame.
    N : np.ndarray of shape (N,)
        Index of refraction for each ray.
    wvl : np.ndarray of shape (N,)
        Wavelength of each ray in microns.
    active : np.ndarray of bool, shape (N,)
        Boolean array indicating if each ray is still active.
    P_hist : list of np.ndarray
        History of positions for each propagation step.
    D_hist : list of np.ndarray
        History of directions for each propagation step.
    normal : np.ndarray of shape (N,3)
        Surface normals at intersection points (updated per surface).
    """

    def __init__(self, P: Union[List[List[float]], np.ndarray],
                       D: Union[List[List[float]], np.ndarray],
                       wvl: Union[float, List[float]] = 0.55,
                       N_0: float = 1.0):
        self.P = np.array(P)  # shape (N,3)
        self.D = np.array(D)  # shape (N,3)
        N_rays = self.P.shape[0]
        # If wvl is a scalar, replicate for all rays.
        if np.isscalar(wvl):
            self.wvl = np.full(N_rays, wvl, dtype=np.float64)
        else:
            self.wvl = np.asarray(wvl, dtype=np.float64)  # type: ignore
        self.N = np.full(N_rays, N_0)
        self.active = np.ones(N_rays, dtype=bool)
        # Check that all direction vectors are normalized.
        norms = np.linalg.norm(self.D, axis=1)
        if not np.allclose(norms, 1.0, atol=0.01):
            raise NormalizationError()
        self.P_hist = [self.P.copy()]
        self.D_hist = [self.D.copy()]
        self.normal = np.zeros_like(self.P)

    def transform(self, surface: geometry) -> None:
        """Transforms positions and directions into the surface's reference frame."""
        self.P = transform_points(surface.R, surface, self.P)
        self.D = transform_dir(surface.R, surface, self.D)

    def find_intersections(self, surface: geometry) -> None:
        """Vectorized computation of intersections of rays with a surface."""
        active_idx = np.where(self.active)[0]
        if active_idx.size == 0:
            return
        # For active rays, compute initial guess for s (parameter along ray)
        P_active = self.P[active_idx]
        D_active = self.D[active_idx]
        # Initial guess: s0 = -P_z / D_z
        s0 = -P_active[:, 2] / D_active[:, 2]
        # Initialize s with s0 instead of zeros
        s = s0.copy()
        error = np.full(s0.shape, 1.0)
        n_iter = 0
        max_iter = int(MAX_INTERSECTION_ITERATIONS)
        tol = INTERSECTION_CONVERGENCE_TOLERANCE
        # Newton-Raphson iteration for all active rays
        while np.any(error > tol) and n_iter < max_iter:
            idx = error > tol
            if not np.any(idx):
                break
            # Update guess points for these rays
            P_temp = P_active[idx] + D_active[idx] * s[idx, None]
            try:
                func, deriv = surface.get_surface_vector(P_temp)
            except Exception:
                # If evaluation fails, mark these rays as inactive
                self.active[active_idx[idx]] = False
                error[idx] = 0.0
                continue
            # Compute derivative along ray direction: dot(deriv, D)
            deriv_dot = np.sum(deriv * D_active[idx], axis=1)
            # Newton update: s_new = s - func / deriv_dot
            delta_s = func / deriv_dot
            s[idx] = s[idx] - delta_s
            error[idx] = np.abs(func)
            n_iter += 1
        # Update positions for active rays based on final s
        P_new = P_active + D_active * s[:, None]
        self.P[active_idx] = P_new
        # Update surface normals for active rays
        try:
            _, norm = surface.get_surface_vector(P_new)
            self.normal[active_idx] = norm
        except Exception:
            self.active[active_idx] = False

    def interact(self, surface: geometry) -> None:
        """Updates ray directions based on interaction type (reflection or refraction).
        
        Uses vectorized operations for active rays.
        """
        active_idx = np.where(self.active)[0]
        if active_idx.size == 0:
            return
        # Determine refractive index ratio mu for active rays.
        if hasattr(surface, 'glass'):
            mu = self.N[active_idx] / surface.glass(self.wvl[active_idx])
        else:
            mu = self.N[active_idx] / surface.N
        # Compute constant a = mu * dot(D, normal) / ||normal||^2
        D_active = self.D[active_idx]
        norm_active = self.normal[active_idx]
        dot_dn = np.sum(D_active * norm_active, axis=1)
        norm_sq = np.sum(norm_active ** 2, axis=1)
        a = mu * dot_dn / norm_sq
        # Compute b = (mu**2 - 1) / norm_sq
        b = (mu**2 - 1) / norm_sq
        if surface.action == 'stop':
            # No change to direction.
            return
        elif surface.action == 'reflection':
            # Reflection: update D = D - 2*a * normal
            self.D[active_idx] = D_active - 2 * a[:, None] * norm_active
        elif surface.action == 'refraction':
            # Identify rays undergoing total internal reflection.
            tir = b > a**2
            if np.any(tir):
                idx_tir = active_idx[tir]
                self.D[idx_tir] = self.D[idx_tir] - 2 * a[tir, None] * norm_active[tir]
            # For rays that are meant to refract:
            ref_idx = active_idx[~tir]
            if ref_idx.size > 0:
                D_ref = self.D[ref_idx]
                norm_ref = self.normal[ref_idx]
                # a_ref and b_ref correspond to the rays that refract.
                a_ref = a[~tir]
                b_ref = b[~tir]
                mu_ref = mu[~tir]  # refractive index ratios for these rays
                # Initial guess for G: -b/(2*a)
                G = -b_ref / (2 * a_ref)
                error = np.full(G.shape, 1.0)
                n_iter = 0
                max_iter = int(MAX_REFRACTION_ITERATIONS)
                tol = REFRACTION_CONVERGENCE_TOLERANCE
                # Newton–Raphson iteration (updating only rays that haven't converged)
                while np.any(error > tol) and n_iter < max_iter:
                    mask = error > tol  # update only nonconverged rays
                    G[mask] = (G[mask]**2 - b_ref[mask]) / (2 * (G[mask] + a_ref[mask]))
                    error[mask] = np.abs(G[mask]**2 + 2 * a_ref[mask] * G[mask] + b_ref[mask])
                    n_iter += 1
                # Identify which rays did not converge
                not_converged = error > tol
                if np.any(not_converged):
                    self.active[ref_idx[not_converged]] = False
                # For rays that converged, update the ray directions and indices.
                converged = ~not_converged
                if np.any(converged):
                    self.D[ref_idx[converged]] = (
                        mu_ref[converged, None] * D_ref[converged]
                        + G[converged, None] * norm_ref[converged]
                    )
                    if hasattr(surface, 'glass'):
                        self.N[ref_idx[converged]] = surface.glass(self.wvl[ref_idx[converged]])
                    else:
                        self.N[ref_idx[converged]] = surface.N

    def ray_lab_frame(self, surface: geometry) -> None:
        """Transforms positions and directions back to the lab frame."""
        self.P = lab_frame_points(surface.R, surface, self.P)
        self.D = lab_frame_dir(surface.R, surface, self.D)

    def update_history(self) -> None:
        """Appends current positions and directions to history arrays."""
        self.P_hist.append(self.P.copy())
        self.D_hist.append(self.D.copy())

    def to_ray_list(self):
        dummy_list = []
        # self.P_hist is a list of arrays [P0, P1, P2, ...]
        # each entry is shape (N, 3) for N rays
        for i in range(self.P.shape[0]):
            dummy = type("DummyRay", (), {})()
            
            # Build the entire per‐surface history for ray i
            dummy.P_hist = [p[i] for p in self.P_hist]
            dummy.D_hist = [d[i] for d in self.D_hist]
            
            # Keep any other attributes needed
            dummy.active = self.active[i]
            dummy.P = self.P[i]
            dummy.D = self.D[i]
            dummy.wvl = self.wvl[i]
            dummy.N = self.N[i]
            dummy.normal = self.normal[i]
            
            dummy_list.append(dummy)
        return dummy_list

    def propagate(self, surfaces: List[geometry]) -> None:
        """Propagates the ray group through a list of surfaces.

        Parameters
        ----------
        surfaces : list of geometry objects
            Surfaces to propagate through in order.
        """
        for surface in surfaces:
            self.transform(surface)
            self.find_intersections(surface)
            if not np.any(self.active):
                break
            self.interact(surface)
            if not np.any(self.active):
                break
            self.ray_lab_frame(surface)
            self.update_history()


import numpy as np

from .ray import RayGroup
from .geometry import geometry

from typing import Union, List, Dict

def ray_plane(geo_params: List[Dict],
              pos: Union[List[float], float, int],
              radius: Union[float, int],
              d: List[float],
              nrays: int = 100,
              wvl: Union[float, int] = 0.55) -> RayGroup:
    """Creates a plane of rays and propagates them through geometry.

    Note
    ----
    This is done by creating a square grid with side length 2*radius
    and then selecting only those points within the circle of given radius.

    Parameters
    ----------
    geo_params : list of dictionaries
        Dictionaries corresponding to surfaces.
    pos : list of length 3 or float/int
        X, Y, and Z values of the plane or just the Z value.
    d : list of length 3
        Direction cosines for each ray.
    radius : float/int
        Radius of ray plane.
    nrays : int
        Approximate number of rays in the ray plane.
    wvl: float/int
        Wavelength of the rays in microns.

    Returns
    -------
    RayGroup
        RayGroup object after propagation through the geometry.
    """
    # Create a square grid and select points within a circle of given radius.
    grid_size = int(4./np.pi*np.sqrt(nrays))  
    x_vals = np.linspace(-radius, radius, grid_size)
    y_vals = np.linspace(-radius, radius, grid_size)
    x_mesh, y_mesh = np.meshgrid(x_vals, y_vals)
    xs = x_mesh.ravel()
    ys = y_mesh.ravel()
    dis = np.sqrt(xs**2 + ys**2)
    mask = dis < radius
    xs = xs[mask]
    ys = ys[mask]
    if isinstance(pos, list):
        x_circ = xs + pos[0]
        y_circ = ys + pos[1]
        z_circ = np.full(xs.shape, pos[2])
    else:
        x_circ = xs
        y_circ = ys
        z_circ = np.full(xs.shape, pos)
    P_arr = np.vstack((x_circ, y_circ, z_circ)).T
    D_arr = np.tile(d, (P_arr.shape[0], 1))
    # Initialize ray group.
    raygroup = RayGroup(P_arr, D_arr, wvl=wvl)
    # Convert geometry dictionaries to geometry objects.
    geo_list = [geometry(surf) for surf in geo_params]
    # Propagate rays through geometry.
    raygroup.propagate(geo_list)
    return raygroup


import tracepy as tp

lens = {
    'action': 'refraction',
    'P': 2.,
    'kappa': -1.25,
    'c': -2,
    'Diam': 2.2
}

back_lens = {
    'action': 'refraction',
    'P': 1.,
    'N': 1.5,
    'Diam': 2.2
}

stop = {
    'action': 'stop',
    'P': 3.,
    'Diam': 2.2
}

def test_rms_hyperbolic():
    geo = [back_lens, lens, stop]
    ray_group = tp.ray_plane(geo, [0., 0., 0.], 1.1, d=[0., 0., 1.], nrays=100)
    rms = tp.spot_rms(geo, ray_group)
    assert rms == 0.


import numpy as np
import tracepy as tp

lens = {
    'name': 'lens',
    'action': 'refraction',
    'P': 2.,
    'kappa': -0.004,
    'c': -0.5,
    'Diam': 2.2
}

back_lens = {
    'name': 'back_lens',
    'action': 'refraction',
    'P': 1.,
    'c': 0.5,
    'kappa': -0.004,
    'N': 1.5,
    'Diam': 2.2
}

stop = {
    'name': 'stop',
    'action': 'stop',
    'P': 3.,
    'Diam': 2.2
}

# Initial geometry
geo_1 = [back_lens, lens, stop]

ray_group = tp.ray_plane(geo_1, [0., 0., 0.], 1.1, d=[0., 0., 1.], nrays=100)

# Define variable elements for optimization
vlens = {
    'name': 'stop',
    'vary': ['P']
}

vlens2 = {
    'name': 'back_lens',
    'vary': ['c', 'kappa', 'N']
}

vlens3 = {
    'name': 'lens',
    'vary': ['c', 'kappa']
}

vary_list = [vlens, vlens2, vlens3]

def test_optimizer():
    try:
        geo_opt = tp.optimize(geo_1, vary_list, typeof='least_squares', max_iter=10)
    except Exception as e:
        assert False, f"Optimization raised an exception: {e}"

    try:
        ray_group_3 = tp.ray_plane(geo_opt, [0., 0., 0.], 1.1, d=[0., 0., 1.], nrays=1000)
        tp.rayaberration(geo_1, ray_group_3)
    except Exception as e:
        assert False, f"rayaberration raised an exception: {e}"


from setuptools import setup

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md')) as f:
    long_description = f.read()

setup(name='tracepy',
      version='0.2.1',
      description='Optical design software for python.',
      url='http://github.com/GNiendorf/tracepy',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Gavin Niendorf',
      author_email='gavinniendorf@gmail.com',
      license='MIT',
      packages=['tracepy'],
      install_requires=[
          'numpy',
          'matplotlib',
          'pandas',
          'scipy',
          'scikit-learn',
          'PyYAML'
      ],
      include_package_data=True,
      )



import pandas as pd

# TODO: Rewrite save_optics to iterate over geo_params to handle edge
#       cases, and convert to dataframe at end before exporting to csv.
def save_optics(geo_params, filename: str):
    """Save geometry to an optics table in csv format.

    Note
    ----
    Directly saves the csv file rather than returning something.

    Parameters
    ----------
    geo_params : list of dictionaries
        Dictionaries correspond to surfaces.
    filename : str
        Output filename.

    """

    table = pd.DataFrame.from_dict(geo_params)
    table[['X', 'Y', 'Z']] = pd.DataFrame(table.P.values.tolist(), index= table.index)
    table[['Alpha', 'Beta', 'Gamma']] = pd.DataFrame(table.D.values.tolist(), index= table.index)
    del[table['P']]; del[table['D']]
    pd.DataFrame.to_csv(table, filename)



import numpy as np

def gen_rot(ang: np.ndarray) -> np.ndarray:
    """Returns a rotation matrix from 3 rotation angles.

    Parameters
    ----------
    ang : np.array of length 3
        Euler angles alpha, beta, gamma in the lab frame.

    Returns
    -------
    np.array((3,3))
        Returns the rotation matrix.

    """

    alpha, beta, gamma = ang
    R_11 = np.cos(alpha)*np.cos(gamma)+np.sin(alpha)*np.sin(beta)*np.sin(gamma)
    R_12 = -np.cos(beta)*np.sin(gamma)
    R_13 = -np.sin(alpha)*np.cos(gamma)+np.cos(alpha)*np.sin(beta)*np.sin(gamma)
    R_21 = np.cos(alpha)*np.sin(gamma)-np.sin(alpha)*np.sin(beta)*np.cos(gamma)
    R_22 = np.cos(beta)*np.cos(gamma)
    R_23 = -np.sin(alpha)*np.sin(gamma)-np.cos(alpha)*np.sin(beta)*np.cos(gamma)
    R_31 = np.sin(alpha)*np.cos(beta)
    R_32 = np.sin(beta)
    R_33 = np.cos(alpha)*np.cos(beta)
    R = np.array([[R_11, R_12, R_13],\
                   [R_21, R_22, R_23],\
                   [R_31, R_32, R_33]])
    return R
